require 'fastcst/ui'
require 'fastcst/repo'
require 'suffix_array'
require 'find'
require 'benchmark'


class IndexCommand < Command
    @@index_list = {}
    
    def initialize(argv)
        @regex = ".*"
        @build = false
        
        super(argv, [
        ["-f", "--files REGEX", "Only search files matching REGEX (default all)", :@regex],
        ["-b", "--build", "Builds the index again", :@regex],
        ])
        
            
        @search = argv
        @repo_dir = Repository::Repository.search
    end
    
    def validate
        valid? @repo_dir, "Could not find repository directory"
        if @repo_dir
            @repo = Repository::Repository.new @repo_dir
        end
            
        # check the regex
        begin
            @regex = Regexp.new @regex
        rescue
            UI.failure :input, $!
            @valid = false
        end
        
        return @valid
    end
    
    # Builds a hash of files->raw_array from SuffixArray.raw_array which
    # is suitable to marshall into a file for later recovery.  Tried using
    # PStore but it had horrible performance for larger sets of data.
    def build_index
        index_db = {}
        
        Find.find(".fastcst/originals") do |file|
            # only index files with something in them
            if File.file?(file) and File.size(file) > 0
                puts "#{file}"
                        
                contents = File.read(file)                        
                sa = SuffixArray.new contents
                        
                array = sa.raw_array
                suffix_start = sa.suffix_start
                        
                index_db[file] = [array, suffix_start]
            end
        end
        
        return index_db
    end
    
    def search_index(index_db, search)
        # this is the first time, so load it into memory for fast searching later
        @@index_list[index_db].each do |file, data|
            if @regex.match(file)
                contents = File.read(file)
                array, suffix_start = data[0], data[1]
                sa = SuffixArray.new contents, array, suffix_start
                start, length = sa.longest_match(search, 0)
                if length == search.length
                    puts "#{file} matched"
                end
            end
        end
    end

    
    def run
        index_file = File.join(@repo.path, "index.db")

        if not File.exist? index_file or @build
            index_db = build_index
            # write the hash just serialzed out
            File.open(index_file, "w") { |f| Marshal.dump(index_db, f) }
            @@index_list["originals"] = index_db
        elsif not @@index_list["originals"]
            # the originals index hasn't been loaded yet
            # snap the whole thing into memory.  doesn't scale but much faster than pstore
            File.open(index_file) { |f| @@index_list["originals"] = Marshal.load(f) }
        end
        
        
        @search.each do |s|
            puts "RESULTS: #{s}"
            search_index("originals", s)
        end
        
    end

end

